@@ -8,7 +8,7 @@ CREATE TABLE nas (
   type varchar(30) DEFAULT 'other',
   ports int(5),
   secret varchar(60) DEFAULT 'secret' NOT NULL,
-#  server varchar(64),
+  server varchar(64),
   community varchar(50),
   description varchar(200) DEFAULT 'RADIUS Client',
   PRIMARY KEY (id),
