@@ -28,6 +28,7 @@ typedef struct listen_detail_t {
 	char		*filename_work;
 	VALUE_PAIR	*vps;
 	FILE		*fp;
+	off_t		offset;
 	detail_state_t 	state;
 	time_t		timestamp;
 	time_t		running;
@@ -36,6 +37,8 @@ typedef struct listen_detail_t {
 	int		signal;
 	int		poll_interval;
 	int		retry_interval;
+	int		packets;
+	int		tries;
 
 	int		has_rtt;
 	int		srtt;
@@ -48,7 +51,7 @@ int detail_recv(rad_listen_t *listener,
 		RAD_REQUEST_FUNP *pfun, REQUEST **prequest);
 int detail_send(rad_listen_t *listener, REQUEST *request);
 void detail_free(rad_listen_t *this);
-int detail_print(rad_listen_t *this, char *buffer, size_t bufsize);
+int detail_print(const rad_listen_t *this, char *buffer, size_t bufsize);
 int detail_encode(UNUSED rad_listen_t *this, UNUSED REQUEST *request);
 int detail_decode(UNUSED rad_listen_t *this, UNUSED REQUEST *request);
 int detail_parse(CONF_SECTION *cs, rad_listen_t *this);
