@@ -66,6 +66,7 @@ struct fr_command_table_t {
 
 typedef struct fr_command_socket_t {
 	char	*path;
+	char	*copy;		/* <sigh> */
 	uid_t	uid;
 	gid_t	gid;
 	int	mode;
@@ -254,6 +255,7 @@ static int fr_server_domain_socket(const char *path)
 	return sockfd;
 }
 
+
 static void command_close_socket(rad_listen_t *this)
 {
 	this->status = RAD_LISTEN_STATUS_CLOSED;
@@ -387,7 +389,7 @@ static void cprint_conf_parser(rad_listen_t *listener, int indent, CONF_SECTION
 			data = variables[i].data;;
 			
 		} else {
-			data = (((char *)base) + variables[i].offset);
+			data = (((const char *)base) + variables[i].offset);
 		}
 
 		switch (variables[i].type) {
@@ -398,7 +400,7 @@ static void cprint_conf_parser(rad_listen_t *listener, int indent, CONF_SECTION
 			
 		case PW_TYPE_INTEGER:
 			cprintf(listener, "%.*s%s = %u\n", indent, tabs,
-				variables[i].name, *(int *) data);
+				variables[i].name, *(const int *) data);
 			break;
 			
 		case PW_TYPE_IPADDR:
@@ -412,7 +414,7 @@ static void cprint_conf_parser(rad_listen_t *listener, int indent, CONF_SECTION
 		case PW_TYPE_BOOLEAN:
 			cprintf(listener, "%.*s%s = %s\n", indent, tabs,
 				variables[i].name, 
-				((*(int *) data) == 0) ? "no" : "yes");
+				((*(const int *) data) == 0) ? "no" : "yes");
 			break;
 			
 		case PW_TYPE_STRING_PTR:
@@ -420,9 +422,9 @@ static void cprint_conf_parser(rad_listen_t *listener, int indent, CONF_SECTION
 			/*
 			 *	FIXME: Escape things in the string!
 			 */
-			if (*(char **) data) {
+			if (*(const char * const *) data) {
 				cprintf(listener, "%.*s%s = \"%s\"\n", indent, tabs,
-					variables[i].name, *(char **) data);
+					variables[i].name, *(const char * const *) data);
 			} else {
 				cprintf(listener, "%.*s%s = \n", indent, tabs,
 					variables[i].name);
@@ -1476,29 +1478,92 @@ static int command_set_module_status(rad_listen_t *listener, int argc, char *arg
 static int command_print_stats(rad_listen_t *listener, fr_stats_t *stats,
 			       int auth)
 {
-	cprintf(listener, "\trequests\t%d\n", stats->total_requests);
-	cprintf(listener, "\tresponses\t%d\n", stats->total_responses);
+	cprintf(listener, "\trequests\t%u\n", stats->total_requests);
+	cprintf(listener, "\tresponses\t%u\n", stats->total_responses);
 	
 	if (auth) {
-		cprintf(listener, "\taccepts\t\t%d\n",
+		cprintf(listener, "\taccepts\t\t%u\n",
 			stats->total_access_accepts);
-		cprintf(listener, "\trejects\t\t%d\n",
+		cprintf(listener, "\trejects\t\t%u\n",
 			stats->total_access_rejects);
-		cprintf(listener, "\tchallenges\t%d\n",
+		cprintf(listener, "\tchallenges\t%u\n",
 			stats->total_access_challenges);
 	}
 
-	cprintf(listener, "\tdup\t\t%d\n", stats->total_dup_requests);
-	cprintf(listener, "\tinvalid\t\t%d\n", stats->total_invalid_requests);
-	cprintf(listener, "\tmalformed\t%d\n", stats->total_malformed_requests);
-	cprintf(listener, "\tbad_signature\t%d\n", stats->total_bad_authenticators);
-	cprintf(listener, "\tdropped\t\t%d\n", stats->total_packets_dropped);
-	cprintf(listener, "\tunknown_types\t%d\n", stats->total_unknown_types);
+	cprintf(listener, "\tdup\t\t%u\n", stats->total_dup_requests);
+	cprintf(listener, "\tinvalid\t\t%u\n", stats->total_invalid_requests);
+	cprintf(listener, "\tmalformed\t%u\n", stats->total_malformed_requests);
+	cprintf(listener, "\tbad_signature\t%u\n", stats->total_bad_authenticators);
+	cprintf(listener, "\tdropped\t\t%u\n", stats->total_packets_dropped);
+	cprintf(listener, "\tunknown_types\t%u\n", stats->total_unknown_types);
 	
 	return 1;
 }
 
 
+#ifdef WITH_DETAIL
+static FR_NAME_NUMBER state_names[] = {
+	{ "unopened", STATE_UNOPENED },
+	{ "unlocked", STATE_UNLOCKED },
+	{ "header", STATE_HEADER },
+	{ "reading", STATE_READING },
+	{ "queued", STATE_QUEUED },
+	{ "running", STATE_RUNNING },
+	{ "no-reply", STATE_NO_REPLY },
+	{ "replied", STATE_REPLIED },
+
+	{ NULL, 0 }
+};
+
+static int command_stats_detail(rad_listen_t *listener, int argc, char *argv[])
+{
+	rad_listen_t *this;
+	listen_detail_t *data;
+	struct stat buf;
+
+	if (argc == 0) {
+		cprintf(listener, "ERROR: Must specify <filename>\n");
+		return 0;
+	}
+
+	data = NULL;
+	for (this = mainconfig.listen; this != NULL; this = this->next) {
+		if (this->type != RAD_LISTEN_DETAIL) continue;
+
+		data = this->data;
+		if (strcmp(argv[1], data->filename) != 0) continue;
+
+		break;
+	}
+
+	cprintf(listener, "\tstate\t%s\n",
+		fr_int2str(state_names, data->state, "?"));
+
+	if ((data->state == STATE_UNOPENED) ||
+	    (data->state == STATE_UNLOCKED)) {
+		return 1;
+	}
+
+	/*
+	 *	Race conditions: file might not exist.
+	 */
+	if (stat(data->filename_work, &buf) < 0) {
+		cprintf(listener, "packets\t0\n");
+		cprintf(listener, "tries\t0\n");
+		cprintf(listener, "offset\t0\n");
+		cprintf(listener, "size\t0\n");
+		return 1;
+	}
+
+	cprintf(listener, "packets\t%d\n", data->packets);
+	cprintf(listener, "tries\t%d\n", data->tries);
+	cprintf(listener, "offset\t%u\n", (unsigned int) data->offset);
+	cprintf(listener, "size\t%u\n", (unsigned int) buf.st_size);
+
+	return 1;
+}
+#endif
+
 #ifdef WITH_PROXY
 static int command_stats_home_server(rad_listen_t *listener, int argc, char *argv[])
 {
@@ -1619,6 +1684,54 @@ static int command_add_client_file(rad_listen_t *listener, int argc, char *argv[
 }
 
 
+static int command_del_client(rad_listen_t *listener, int argc, char *argv[])
+{
+#ifdef WITH_DYNAMIC_CLIENTS
+	RADCLIENT *client;
+
+	client = get_client(listener, argc - 1, argv + 1);
+	if (!client) return 0;
+
+	if (!client->dynamic) {
+		cprintf(listener, "ERROR: Client %s was not dynamically defined.\n", argv[1]);
+		return 0;
+	}
+
+	/*
+	 *	DON'T delete it.  Instead, mark it as "dead now".  The
+	 *	next time we receive a packet for the client, it will
+	 *	be deleted.
+	 *
+	 *	If we don't receive a packet from it, the client
+	 *	structure will stick around for a while.  Oh well...
+	 */
+	client->lifetime = 1;
+#else
+	cprintf(listener, "ERROR: Dynamic clients are not supported.\n");
+#endif
+
+	return 1;
+}
+
+
+static fr_command_table_t command_table_del_client[] = {
+	{ "ipaddr", FR_WRITE,
+	  "del client ipaddr <ipaddr> - Delete a dynamically created client",
+	  command_del_client, NULL },
+
+	{ NULL, 0, NULL, NULL, NULL }
+};
+
+
+static fr_command_table_t command_table_del[] = {
+	{ "client", FR_WRITE,
+	  "del client <command> - Delete client configuration commands",
+	  NULL, command_table_del_client },
+
+	{ NULL, 0, NULL, NULL, NULL }
+};
+
+
 static fr_command_table_t command_table_add_client[] = {
 	{ "file", FR_WRITE,
 	  "add client file <filename> - Add new client definition from <filename>",
@@ -1684,6 +1797,12 @@ static fr_command_table_t command_table_stats[] = {
 	  command_stats_home_server, NULL },
 #endif
 
+#ifdef WITH_DETAIL
+	{ "detail", FR_READ,
+	  "stats detail <filename> - show statistics for the given detail file",
+	  command_stats_detail, NULL },
+#endif
+
 	{ NULL, 0, NULL, NULL, NULL }
 };
 
@@ -1692,6 +1811,7 @@ static fr_command_table_t command_table[] = {
 	{ "debug", FR_WRITE,
 	  "debug <command> - debugging commands",
 	  NULL, command_table_debug },
+	{ "del", FR_WRITE, NULL, NULL, command_table_del },
 	{ "hup", FR_WRITE,
 	  "hup [module] - sends a HUP signal to the server, or optionally to one module",
 	  command_hup, NULL },
@@ -1712,6 +1832,16 @@ static fr_command_table_t command_table[] = {
 };
 
 
+static void command_socket_free(rad_listen_t *this)
+{
+	fr_command_socket_t *sock = this->data;
+
+	unlink(sock->copy);
+	free(sock->copy);
+	sock->copy = NULL;
+}
+
+
 /*
  *	Parse the unix domain sockets.
  *
@@ -1721,12 +1851,17 @@ static int command_socket_parse(CONF_SECTION *cs, rad_listen_t *this)
 {
 	fr_command_socket_t *sock;
 
+	if (check_config) return 0;
+
 	sock = this->data;
 
 	if (cf_section_parse(cs, sock, command_config) < 0) {
 		return -1;
 	}
 
+	sock->copy = NULL;
+	if (sock->path) sock->copy = strdup(sock->path);
+
 #if defined(HAVE_GETPEEREID) || defined (SO_PEERCRED)
 	if (sock->uid_name) {
 		struct passwd *pw;
@@ -1786,7 +1921,7 @@ static int command_socket_parse(CONF_SECTION *cs, rad_listen_t *this)
 	return 0;
 }
 
-static int command_socket_print(rad_listen_t *this, char *buffer, size_t bufsize)
+static int command_socket_print(const rad_listen_t *this, char *buffer, size_t bufsize)
 {
 	fr_command_socket_t *sock = this->data;
 
@@ -2029,7 +2164,7 @@ static int command_domain_recv(rad_listen_t *listener,
 			 */
 			if (((co->mode & FR_WRITE) == 0) &&
 			    ((table[i].mode & FR_WRITE) != 0)) {
-				cprintf(listener, "ERROR: You do not have write permission.  See \"mode = rw\" in %s\n", co->path);
+				cprintf(listener, "ERROR: You do not have write permission.  See \"mode = rw\" in the \"listen\" section for this socket.\n");
 				goto do_next;
 			}
 
