@@ -1930,7 +1930,7 @@ static modcallable *do_compile_modsingle(modcallable *parent,
 		}
 		
 		*modname = NULL;
-		cf_log_err(ci, "Failed to find module \"%s\".", modrefname);
+		cf_log_err(ci, "Failed to load module \"%s\".", modrefname);
 		return NULL;
 	} while (0);
 
