@@ -49,6 +49,8 @@ static const CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,sql_password), NULL, ""},
 	{"radius_db", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,sql_db), NULL, "radius"},
+	{"filename", PW_TYPE_FILENAME, /* for sqlite */
+	 offsetof(SQL_CONFIG,sql_file), NULL, NULL},
 	{"read_groups", PW_TYPE_BOOLEAN,
 	 offsetof(SQL_CONFIG,read_groups), NULL, "yes"},
 	{"sqltrace", PW_TYPE_BOOLEAN,
@@ -398,12 +400,13 @@ static int generate_sql_clients(SQL_INST *inst)
 			c->nastype = strdup(row[3]);
 
 		numf = (inst->module->sql_num_fields)(sqlsocket, inst->config);
-		if ((numf > 5) && (row[5] != NULL)) c->server = strdup(row[5]);
+		if ((numf > 5) && (row[5] != NULL) && *row[5]) c->server = strdup(row[5]);
 
 		DEBUG("rlm_sql (%s): Adding client %s (%s, server=%s) to clients list",
 		      inst->config->xlat_name,
 		      c->longname,c->shortname, c->server ? c->server : "<none>");
 		if (!client_add(NULL, c)) {
+			sql_release_socket(inst, sqlsocket);
 			DEBUG("rlm_sql (%s): Failed to add client %s (%s) to clients list.  Maybe there's a duplicate?",
 			      inst->config->xlat_name,
 			      c->longname,c->shortname);
@@ -880,8 +883,38 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 	}
 
 	xlat_name = cf_section_name2(conf);
-	if (xlat_name == NULL)
+	if (xlat_name == NULL) {
 		xlat_name = cf_section_name1(conf);
+	} else {
+		char *group_name;
+		DICT_ATTR *dattr;
+		ATTR_FLAGS flags;
+
+		/*
+		 * Allocate room for <instance>-SQL-Group
+		 */
+		group_name = rad_malloc((strlen(xlat_name) + 1 + 11) * sizeof(char));
+		sprintf(group_name,"%s-SQL-Group",xlat_name);
+		DEBUG("rlm_sql Creating new attribute %s",group_name);
+
+		memset(&flags, 0, sizeof(flags));
+		dict_addattr(group_name, 0, PW_TYPE_STRING, -1, flags);
+		dattr = dict_attrbyname(group_name);
+		if (dattr == NULL){
+			radlog(L_ERR, "rlm_ldap: Failed to create attribute %s",group_name);
+			free(group_name);
+			free(inst);	/* FIXME: detach */
+			return -1;
+		}
+
+		if (inst->config->groupmemb_query && 
+		    inst->config->groupmemb_query[0]) {
+			DEBUG("rlm_sql: Registering sql_groupcmp for %s",group_name);
+			paircompare_register(dattr->attr, PW_USER_NAME, sql_groupcmp, inst);
+		}
+
+		free(group_name);
+	}
 	if (xlat_name){
 		inst->config->xlat_name = strdup(xlat_name);
 		xlat_register(xlat_name, (RAD_XLAT_FUNC)sql_xlat, inst);
@@ -936,7 +969,10 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 		return -1;
 	}
 
-	paircompare_register(PW_SQL_GROUP, PW_USER_NAME, sql_groupcmp, inst);
+	if (inst->config->groupmemb_query && 
+	    inst->config->groupmemb_query[0]) {
+		paircompare_register(PW_SQL_GROUP, PW_USER_NAME, sql_groupcmp, inst);
+	}
 
 	if (inst->config->do_clients){
 		if (generate_sql_clients(inst) == -1){
